# -*- coding: utf-8 -*-
"""Wavelet manipulation.

author:
    Leonard Seydoux and Randall Balestriero
"""

import numpy as cp
import numpy as np

from scipy.signal import tukey


def gaussian_window(x, width):
    """Gaussian window.

    This function can generate a bank of windows at once if the width
    argument is a vector (and/or amplitude). In this case, it should have
    a new axis with respect to the time vector to allow for outer product.

    Parameters
    ----------
    x : :class:`T.ndarray` or np.ndarray
        Input variable (in the same units than the width).
    width : float or np.ndarray
        Window width (in the same units than the input variable). If an array
        is provided, the function returns as many windows as the number of
        elements of this parameter.
    amplitude : float or np.ndarray, optional
        Window amplitude at maximum (default 1). If this parameter is a vector,
        it should have the same number of elements than the width.

    Returns
    -------
    :class:`T.ndarray`
        The Gaussian window in the time domain. If the width (and possibly
        amplitude) argument is a vector, the function returns a matrix with
        shape (len(width), len(x)).
    """
    # turn parameters into a numpy arrays for dimension check
    x = cp.array(x)
    width = cp.array(width)

    # add new axis for outer product if several widths are given
    width = width[:, None] if width.shape and (width.ndim == 1) else width

    return cp.exp(-((x / width) ** 2))


def complex_morlet(x, center, width):
    """Complex Morlet wavelet.

    The complex Morlet wavelet is a complex plane wave modulated by a
    Gaussian window. The oscillatory frequency of the plane wave is the
    center frequency, and the temporal width of the Gaussian is the width
    argument.

    This function can generate a filter bank at once if the width and center
    arguments are vectors of the same size. In this case, they should have a
    new axis with respect to the time vector to allow for outer product.

    Arguments
    ---------
    x: :class:`T.ndarray` or np.ndarray
        Time vector in seconds.

    width: float or :class:`T.ndarray` or np.ndarray
        Temporal signal width in seconds.

    center: float or :class:`T.ndarray` or np.ndarray
        Center frequency in hertz.

    Keyword arguments
    -----------------
    amplitude: float (optional)
        Wavelet normalization (default 1). If amplitude is a vector, it should
        have the same dimension than width (and center).

    Returns
    -------
    filter: :class:`T.ndarray`
        The complex Mortlet wavelet in the time domain. If the center and width
        (and possibly amplitude) arguments are vectors, the function returns
        a matrix with shape (len(width), len(x)).
    """
    # turn parameters into a numpy arrays for dimension check
    x = cp.array(x)
    width = cp.array(width)
    center = cp.array(center)

    # add new axis for outer product if several widths are given
    width = width[:, None] if width.shape else width
    center = center[:, None] if center.shape else center

    # check compatibility between arguments
    if width.shape and center.shape:
        assert (
            width.shape == center.shape
        ), f"Shape for widths {width.shape} and centers {center.shape} differ."

    return gaussian_window(x, width) * cp.exp(2j * cp.pi * center * x)


class ComplexMorletBank:
    """Complex Morlet filter bank."""

    def __init__(
        self, bins, octaves, resolution=1, quality=4, taper_alpha=1e-3
    ):
        """Filter bank creation.

        This function creates the filter bank in the time domain, and obtains
        it in the frequency domain with a fast Fourier transform.

        Arguments
        ---------
        bins: int
            Number of samples in the time domain.

        octaves: int
            Number of octaves spanned by the filter bank.

        Keyword arguments
        -----------------
        resolution: int
            Number of filters per octaves (default 1).

        sampling: float
            Input data sampling rate (default 1 Hz).

        quality: float
            Filter bank quality factor (constant, default 4).

        """
        # attribution
        self.bins = bins
        self.octaves = octaves
        self.resolution = resolution
        self.quality = quality

        # generate bank
        self.wavelets = complex_morlet(
            self.times(), self.centers(), self.widths()
        )
        self.spectra = cp.fft.fft(self.wavelets)
        self.size = self.wavelets.shape[0]
        self.taper = cp.array(tukey(bins, alpha=taper_alpha))
        pass

    def transform(self, sample):
        """Scalogram applied to a data sample.

        Arguments
        ---------
        x: np.ndarray
            A data sample of shape `(..., channels, bins)`, with the same
            number of bins than the filter bank.

        Returns
        -------
        wx: cp.ndarray
            The scalograms for all channels with shape (the ellipsis stands for
            unknown number of input dimensions)
            `n_channels, ..., n_filters, n_bins`.
        """
        sample = cp.fft.fft(cp.array(sample) * self.taper)
        convolved = sample[..., None, :] * self.spectra
        scalogram = cp.fft.fftshift(cp.fft.ifft(convolved), axes=-1)
        return cp.abs(scalogram)

    def times(self, sampling_rate=1):
        """Wavelet bank symmetric time vector in seconds."""
        duration = self.bins / sampling_rate
        return np.linspace(-0.5, 0.5, num=self.bins) * duration

    def frequencies(self, sampling_rate=1):
        """Wavelet bank frequency vector in hertz."""
        return np.linspace(0, sampling_rate, self.bins)

    def nyquist(self, sampling_rate=1):
        """Wavelet bank frequency vector in hertz."""
        return sampling_rate / 2

    @property
    def shape(self):
        """Filter bank total number of filters."""
        return self.octaves * self.resolution, self.bins

    @property
    def ratios(self):
        """Wavelet bank ratios."""
        ratios = np.linspace(self.octaves, 0.0, self.shape[0], endpoint=False)
        return -ratios[::-1]

    @property
    def scales(self):
        """Wavelet bank scaling factors."""
        return 2 ** self.ratios

    def centers(self, sampling_rate=1):
        """Wavelet bank center frequencies."""
        return self.scales * self.nyquist(sampling_rate)

    def widths(self, sampling_rate=1):
        """Wavelet bank temporal widths."""
        return self.quality / self.centers(sampling_rate)




# coding: utf-8
"""Command-line module for data inventory.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""

import click
import os
import pickle

from .common import common_options
from .transform import load_waveform

from ..io import mkdir, stdout
from ..display import show_waveforms


@click.command("waveforms", short_help="Show waveforms from clustering.")
@common_options
@click.option(
    "--segment",
    type=float,
    default=200.0,
    show_default=True,
    help="Segment duration (seconds).",
)
@click.option(
    "--n_samples",
    type=int,
    default=10,
    show_default=True,
    help="Number of waveforms to show.",
)
@click.option(
    "--dimensions",
    default=10,
    show_default=True,
    help="Number of latent space dimensions.",
    type=int,
)
@click.option(
    "--normalize",
    is_flag=True,
    default=False,
    show_default=True,
    help="Normalization flag.",
)
def waveforms(
    segment,
    dimensions,
    normalize,
    n_samples,
    savepath=None,
    figpath=None,
    filename_inventory=None,
    filename_network=None,
    path_features=None,
    path_clusters=None,
    **kwargs,
):
    """Transform seismograms into scattering domain."""
    # Path
    dirpath_clusters = os.path.join(savepath, path_clusters)
    dirpath_features = os.path.join(savepath, path_features)
    dirpath_inventory = os.path.join(savepath, "inventories")
    dirpath_figure = os.path.join(figpath, "waveforms", "waveforms_{}_{}")

    # Files
    filepath_features = os.path.join(dirpath_features, "features_{}_{}.npz")
    filepath_clusters = os.path.join(dirpath_clusters, "clusters_{}_{}.npz")
    filepath_inventory = os.path.join(dirpath_inventory, filename_inventory)
    filepath_figure = os.path.join(dirpath_figure, "cluster_*")

    # Append
    norm = "norm" if normalize is True else "no-norm"
    dirpath_figure = dirpath_figure.format(dimensions, norm)
    filepath_features = filepath_features.format(dimensions, norm)
    filepath_clusters = filepath_clusters.format(dimensions, norm)
    filepath_figure = filepath_figure.format(dimensions, norm)

    mkdir(dirpath_figure)
    show_waveforms(
        segment,
        filepath_features,
        filepath_clusters,
        filepath_inventory,
        filepath_figure,
        n_samples=n_samples,
        reader=load_waveform,
        factor=0.8,
    )
    pass



# coding=utf-8
"""Deep scattering transform clustering on segmented time series."""

__all__ = [
    "display",
    "parse",
    "Network",
    "inventory",
    "signal",
    "wavelet",
    "hierarchy",
]

from . import display
from . import wavelet
from . import inventory
from . import signal
from . import hierarchy
from .io import parse
from .network import Network

from pkg_resources import get_distribution

__version__ = get_distribution('scatseisnet').version


# coding: utf-8
"""Calculate scattering transform on segmented time series."""

import click
import numpy as np
import os
import pickle

from dateutil import tz
from matplotlib import dates as mdates

from .common import common_options

from .. import hierarchy
from ..display import dendrogram
from ..io import stdout, mkdir


@click.command("linkage", short_help="Calculate linkage matrix.")
@common_options
@click.option(
    "--method",
    type=click.Choice(("single", "centroid", "median", "ward")),
    default="ward",
    help="Number of clusters splits.",
)
@click.option(
    "--dimensions",
    default=10,
    show_default=True,
    help="Number of latent space dimensions.",
    type=int,
)
@click.option(
    "--normalize",
    is_flag=True,
    default=False,
    show_default=True,
    help="Normalization flag.",
)
@click.option(
    "--n_clusters",
    "-n",
    type=int,
    default=10,
    show_default=True,
    help="Number of clusters.",
)
@click.option(
    "--time-zone",
    default="Mexico/General",
    type=str,
    show_default=True,
    help="Time zone for local time histogram.",
)
def linkage(
    method,
    n_clusters,
    dimensions,
    normalize,
    time_zone,
    path_features=None,
    path_clusters=None,
    savepath=None,
    figpath=None,
    show=False,
    **kwargs,
):
    # Path
    dirpath_clusters = os.path.join(savepath, path_clusters)
    dirpath_features = os.path.join(savepath, path_features)

    # Files
    filepath_features = os.path.join(dirpath_features, "features_{}_{}.npz")
    filepath_linkage = os.path.join(dirpath_clusters, "linkage_{}_{}.npz")
    filepath_clusters = os.path.join(dirpath_clusters, "clusters_{}_{}.npz")
    filepath_dendrogram = os.path.join(figpath, "dendrogram_{}_{}.png")

    # Append
    norm = "norm" if normalize is True else "no-norm"
    filepath_features = filepath_features.format(dimensions, norm)
    filepath_linkage = filepath_linkage.format(dimensions, norm)
    filepath_clusters = filepath_clusters.format(dimensions, norm)
    filepath_dendrogram = filepath_dendrogram.format(dimensions, norm)

    if show is True:
        mkdir(figpath)
        linkage = pickle.load(open(filepath_linkage, "rb"))
        stdout("Loaded linkage matrix from", filepath_linkage)
        with np.load(filepath_features) as data:
            timestamps = data["times"]
        times = np.array(mdates.num2date(timestamps, tz=tz.gettz(time_zone)))

        # Show dendrogram
        fig, predictions = dendrogram(linkage, times, n_clusters)
        fig.savefig(filepath_dendrogram)
        stdout("Saved figure at", filepath_dendrogram)

        # Save predictions
        np.savez(filepath_clusters, predictions=predictions, times=timestamps)
        stdout("Saved predictions at", filepath_clusters)

    else:
        mkdir(dirpath_clusters)

        # Load features
        with np.load(filepath_features) as data:
            features = data["features"]
        stdout("Features loaded from", filepath_features)

        # Calculate and save linkage
        linkage = hierarchy.linkage(features, method)
        pickle.dump(linkage, open(filepath_linkage, "wb"))
        stdout("Saved linkage matrix at", filepath_linkage)



# coding: utf-8
"""Display features.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""
import numpy as np

from matplotlib import dates as mdates
from matplotlib import pyplot as plt
from string import ascii_lowercase as letters
from scipy.cluster import hierarchy


COLORS = [
    "0.8",
    "#222222",
    "#F3C300",
    "#875692",
    "#F38400",
    "#A1CAF1",
    "#BE0032",
    "#C2B280",
    "#848482",
    "#008856",
    "#E68FAC",
    "#0067A5",
    "#F99379",
    "#604E97",
    "#F6A600",
    "#B3446C",
    "#DCD300",
    "#882D17",
    "#8DB600",
    "#654522",
    "#E25822",
    "#2B3D26",
]


def get_leaves(dendrogram_info, ax):
    """Get dendrogram list of leaves coordinates and colors.

    Arguments
    ---------
    d: dict
        Output of the scipy.hierarchy.dendrogram function.

    Returns
    -------
    coords, colors: array-like
        The x coordinates and color of each leave.
    """
    # Extract coordinates of each leave (with a depth coordinate indexed 0)
    infos = (key for key in dendrogram_info)
    node_index, node_depth, *_ = (dendrogram_info[key] for key in infos)
    leaves_coordinates = list()
    for index, depth in zip(node_index, node_depth):
        if depth[0] == 0:
            leaves_coordinates.append(index[0])
        if depth[-1] == 0:
            leaves_coordinates.append(index[-1])
    leaves_coordinates = sorted(set(leaves_coordinates))

    # Cardinality
    leaves_population_size = list()
    for label in ax.get_yticklabels():
        label = label.get_text()
        label = label.replace("(", "").replace(")", "")
        population_size = int(label)
        leaves_population_size.append(population_size)

    return leaves_coordinates, leaves_population_size


def get_prediction(linkage, population_size):
    """Get cluster predection for each sample.

    Arguments
    ---------
    linkage: np.ndarray
        Output of the linkage function.
    population_size: list
        The size of every cluster
    """
    indexes_flat = hierarchy.leaves_list(linkage)
    predictions = np.zeros_like(indexes_flat)
    start = 0
    for index, size in enumerate(population_size):
        predictions[indexes_flat[start : start + size]] = index + 1
        start += size
    return predictions


def show_dendrogram(linkage, ax=plt.gca(), depth=30):
    """Show dendrogram and returns basic cluster informations.

    Arguments
    ---------
    linkage: np.ndarray
        Linkage matrix.

    Keyword arguments
    -----------------
    ax: plt.Axes
        The axes to draw the dendrogram into.
    depth: str
        The dendrogram depth

    Return
    ------
    prediction: np.ndarray
        The cluster prediction per sample.
    """
    # Show and get dendrogram
    with plt.rc_context({"lines.linewidth": 0.7}):
        dendrogram_infos = hierarchy.dendrogram(
            linkage,
            p=depth,
            truncate_mode="lastp",
            color_threshold=0,
            ax=ax,
            orientation="left",
            above_threshold_color="0.3",
            count_sort=True,
            labels=None,
        )

    # Extract informations
    coordinates, population_sizes = get_leaves(dendrogram_infos, ax)
    predictions = get_prediction(linkage, population_sizes)

    # Plot leave nodes
    node_style = dict(ms=5, mec="0.3", mew=0.7, clip_on=False)
    for coordinate, color in zip(coordinates, COLORS):
        ax.plot(0, coordinate, "o", mfc=color, **node_style)
        index = int((coordinate - 5) / 10) + 1
        label = "{:d}".format(index)
        ax.text(-0.1, coordinate, label, color=color, va="center")

    return predictions


def dendrogram(
    linkage, times, n_clusters, hourly=np.arange(24), n_cal_bins=150
):

    # Deactivate axes basic properties
    spines_off = {
        "axes.spines.right": False,
        "axes.spines.left": False,
        "axes.spines.top": False,
        "axes.facecolor": "none",
        "xtick.top": False,
        "ytick.left": False,
    }

    # Generate axes
    gs = dict(width_ratios=[2, 4, 1, 2])
    figsize = 6, n_clusters * 0.35
    with plt.rc_context(spines_off):
        figure_kwargs = dict(sharey=True, figsize=figsize, gridspec_kw=gs)
        figure, axes = plt.subplots(1, 4, **figure_kwargs)

    # Axes unpack
    ax_dendrogram, ax_cal, ax_hourly, ax_population = axes

    # Calendar bins
    timestamps = mdates.date2num(times)
    edge_shift = 0.1 * (timestamps[-1] - timestamps[0])
    start, end = timestamps[0] - edge_shift, timestamps[-1] + edge_shift
    cal_bins = np.linspace(start, end, n_cal_bins)
    cal_step = cal_bins[1] - cal_bins[0]
    h_step = hourly[1] - hourly[0]

    # Show dendrogram
    predictions = show_dendrogram(linkage, ax=ax_dendrogram, depth=n_clusters)
    classes = sorted(set(predictions))

    # Show other cluster properties
    for cluster, color in zip(classes, COLORS):

        # Cluster coordinates
        yshift = (cluster - 1) * 10 + 5
        indexes = predictions == cluster

        # Population size
        size = np.sum(predictions == cluster)
        ratio = 100 * size / len(times)

        # Calendar occurrences
        cluster_times = times[indexes]
        cluster_timestamps = timestamps[indexes]
        cluster_hours = [time.hour for time in cluster_times]
        cal_counts, _ = np.histogram(cluster_timestamps, cal_bins)
        cal_counts = cal_counts / cal_counts.max()

        # Hourly occurrences
        hourly_counts = np.sum(cluster_hours == hourly[:, None], axis=1)
        hourly_counts = hourly_counts / hourly_counts.max()

        # Population graph
        bar_style = dict(height=3, color=color, ec="0.3", lw=0.5, align="edge")
        text_style = dict(size=6, va="center", color=color)
        text_label = f" {size}"
        ax_population.barh(yshift, ratio, **bar_style)
        ax_population.text(ratio, yshift + 1.5, text_label, **text_style)

        # Calendar graph
        bar_style = dict(bottom=yshift, width=cal_step, fc=color, align="edge")
        step_style = dict(c="0.3", lw=0.5, where="post")
        ax_cal.bar(cal_bins[:-1], cal_counts * 5, **bar_style)
        ax_cal.step(cal_bins[:-1], cal_counts * 5 + yshift, **step_style)

        # Hourly graph
        bar_style = dict(bottom=yshift, width=h_step, fc=color, align="edge")
        step_style = dict(c="0.3", lw=0.5, where="post")
        ax_hourly.bar(hourly, hourly_counts * 5, **bar_style)
        ax_hourly.step(hourly, hourly_counts * 5 + yshift, **step_style)

    # Labels dendrogram
    ax_dendrogram.set_xlabel("Rescaled distance")
    ax_dendrogram.set_yticklabels([])
    ax_dendrogram.yaxis.set_label_position("right")

    # Labels population
    ax_population.set_yticks(10 * np.arange(len(classes)) + 5)
    ax_population.set_xlabel("Relative\npopulation size (%)", loc="left")

    # Labels calendar
    ax_cal.set_yticks(10 * np.arange(len(classes)) + 5)
    ax_cal.set_xlabel("Calendar date", loc="left")
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks, show_offset=False)
    ax_cal.xaxis.set_major_locator(dateticks)
    ax_cal.xaxis.set_major_formatter(datelabels)
    ax_cal.set_xlim(start, end)
    plt.setp(ax_cal.get_xticklabels(), rotation="vertical")

    # Labels hourly
    hours_ticks = range(0, 25, 12)
    hours_labels = [f"{h:02d}" for h in hours_ticks]
    ax_hourly.set_xlim(0, 24)
    ax_hourly.set_xlabel("Local\ntime (hours)", loc="left")
    ax_hourly.set_xticks(hours_ticks)
    ax_hourly.set_xticklabels(hours_labels)

    # All-axes cosmetics
    for ax, letter in zip(axes, letters):
        ax.grid(clip_on=False)
        ax.set_title(letter, loc="left")

    return figure, predictions



# coding: utf-8
"""Display inventory elements.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""

import numpy as np

from matplotlib import dates as mdates
from matplotlib import pyplot as plt
from matplotlib import ticker
from string import ascii_lowercase as letters

# from ..inventory import read
from ..io import stdout

plt.rcParams["figure.figsize"] = 5, 5
plt.rcParams["figure.constrained_layout.use"] = True


def show_availability(start, end, ax=plt.gca()):
    """Show the daily data availability along calendar.
    
    Arguments
    ---------
    start: array-like
        Series of starttimes for each segment.
    end: array-like
        Series of endtimes for each segment. Should have same number of
        elements than start.
    ax: plt.Axes, optional
        Axes to use for drawing.
    """
    # Subproducts
    duration_days = end - start
    duration_hour = duration_days * 24
    start_hour = (start - np.fix(start)) * 24

    # Plot
    ax.bar(
        start,
        duration_hour,
        width=duration_days,
        bottom=start_hour,
        align="edge",
        label="Available",
    )

    # Normal duration
    ax.axhspan(0, 24, fc="C0", alpha=0.1, zorder=0, label="Normal")

    # Labels
    ax.legend()
    ax.set_xlabel("Calendar date")
    ax.set_ylabel("Time (H:M)")
    ax.grid(which="both")

    # Y-axis
    y_min = min(0, np.min(start_hour))
    y_max = max(24, np.max(start_hour + duration_hour))
    ax.set_ylim(y_min, y_max)
    ax.yaxis.set_major_locator(ticker.MultipleLocator(base=6))
    ax.yaxis.set_major_formatter(ticker.FormatStrFormatter("%02d:00"))

    # X-axis
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    ax.xaxis.set_major_locator(dateticks)
    ax.xaxis.set_major_formatter(datelabels)

    pass


def show_gaps(start, end, ax=plt.gca()):
    """Show the daily data gaps along calendar.
    
    Arguments
    ---------
    start: array-like
        Series of starttimes for each segment.
    end: array-like
        Series of endtimes for each segment. Should have same number of
        elements than start.
    ax: plt.Axes, optional
        Axes to use for drawing.
    """
    gaps = (start[1:] - end[:-1]).clip(0, None)

    # Show
    ax.bar(
        end[:-1],
        gaps,
        width=gaps,
        bottom=0.1,
        align="edge",
        facecolor="C1",
        label="Gap",
    )

    # Labels
    ax.legend()
    ax.set_xlabel("Calendar date")
    ax.set_ylabel("Gap duration (days)")
    ax.grid(which="both")

    # Indicate typical durations
    x_max = ax.get_xlim()[1]
    y_max = ax.get_ylim()[1]
    durations = {" day": 1, " week": 7, " month": 30, " year": 365}
    text_style = {"va": "center", "color": "0.3"}
    for duration_name, duration_days in durations.items():
        if duration_days < y_max:
            ax.axhline(duration_days, c="0.3", ls="--")
            ax.text(x_max, duration_days, duration_name, **text_style)

    # Y-axis
    ax.set_yscale("log")
    ax.set_ylim(bottom=0.1)

    # X-axis
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    ax.xaxis.set_major_locator(dateticks)
    ax.xaxis.set_major_formatter(datelabels)


def show_sampling_rate(start, sampling_rate, ax=plt.gca()):
    """Show the number of zeros in segments.
    
    Arguments
    ---------
    start: array-like
        Series of starttimes for each segment.
    sampling_rate: array-like
        Sampling rate for each segment. Should have same number of
        elements than start.
    ax: plt.Axes, optional
        Axes to use for drawing.
    """
    # Show
    ax.step(start, sampling_rate, where="post", c="C3")
    ax.fill_between(start, 0, sampling_rate, step="post", fc="C3", alpha=0.2)

    # Labels
    ax.set_ylabel("Sampling rate (Hz)")
    ax.set_xlabel("Calendar date")
    ax.grid(which="both")

    # Y-axis
    ax.set_ylim(bottom=0)

    # X-axis
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    ax.xaxis.set_major_locator(dateticks)
    ax.xaxis.set_major_formatter(datelabels)


def show_zeros(start, zeros, ax=plt.gca()):
    """Show the number of zeros in segments.
    
    Arguments
    ---------
    start: array-like
        Series of starttimes for each segment.
    zeros: array-like
        Number of zeros for each segment. Should have same number of
        elements than start.
    ax: plt.Axes, optional
        Axes to use for drawing.
    """
    # Show
    ax.step(start, zeros, where="post", c="C2")
    ax.fill_between(start, 0, zeros, step="post", fc="C2", alpha=0.2)

    # Labels
    ax.set_ylabel("Number of zeros")
    ax.set_xlabel("Calendar date")
    ax.grid(which="both")

    # Y-axis
    ax.set_yscale("log")

    # X-axis
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    ax.xaxis.set_major_locator(dateticks)
    ax.xaxis.set_major_formatter(datelabels)


def show_npts(start, npts, ax=plt.gca()):
    """Show the number of zeros in segments.
    
    Arguments
    ---------
    start: array-like
        Series of starttimes for each segment.
    zeros: array-like
        Number of zeros for each segment. Should have same number of
        elements than start.
    ax: plt.Axes, optional
        Axes to use for drawing.
    """
    # Show
    npts = np.array(npts)
    for npt in npts.T:
        ax.step(start, npt, where="post", c="C5")
        ax.fill_between(start, 0, npt, step="post", fc="C5", alpha=0.2)

    # Labels
    ax.set_ylabel("Number of points")
    ax.set_xlabel("Calendar date")
    ax.grid(which="both")

    # Y-axis
    ax.set_yscale("log")

    # X-axis
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    ax.xaxis.set_major_locator(dateticks)
    ax.xaxis.set_major_formatter(datelabels)


def show_main(start, end, zeros, sampling_rate, npts):
    """Show different elements of inventory."""
    # Create figure
    figure, axes = plt.subplots(5, sharex=True, figsize=(5, 6))

    # Plot
    show_availability(start, end, axes[0])
    show_gaps(start, end, axes[1])
    show_zeros(start, zeros, axes[2])
    show_sampling_rate(start, sampling_rate, axes[3])
    show_npts(start, npts, axes[4])

    # Common labels
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    for letter, axe in zip(letters, axes):
        axe.xaxis.set_major_locator(dateticks)
        axe.xaxis.set_major_formatter(datelabels)
        axe.set_xlabel("")
        axe.set_title(letter, loc="left")
    axes[-1].set_xlabel("Calendar date", loc="left")

    return figure


def get_all_subkeys(dictionnary, subkey):
    """Get all sub entries of a dictionnary with given subkey."""
    subvalues = [dictionnary[key][subkey] for key in dictionnary.keys()]
    return subvalues


def show_inventory(database, file_figure):
    """Show inventory and save figure.
    
    Arguments
    ---------
    database: dict
        The inventory.
    file_figure:
        The figure file.
    """
    start = get_all_subkeys(database, "date_start")
    end = get_all_subkeys(database, "date_end")
    zeros = get_all_subkeys(database, "n_zeros")
    sampling_rate = get_all_subkeys(database, "sampling_rate")
    npts = get_all_subkeys(database, "n_pts")
    start = mdates.date2num(start)
    end = mdates.date2num(end)

    # Make and save figure
    fig = show_main(start, end, zeros, sampling_rate, npts)
    fig.savefig(file_figure)
    stdout("Figure saved at", file_figure)




# coding: utf-8
"""Display features.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""
import nmmn.plots
import numpy as np
import pickle

from matplotlib import dates as mdates
from matplotlib import pyplot as plt

from .. import inventory
from .. import signal
from ..io import load_feature_file, stdout

plt.rcParams["figure.constrained_layout.use"] = True


def show(t, x, sx, net, timestamp, channel=1):

    # Create figure
    fig, ax = plt.subplots(3, figsize=(5, 6))

    # Metadata
    sampling_rate = net.sampling_rate
    freq = [bank.centers(sampling_rate) for bank in net.banks]

    # Log-transformatino
    sx = np.log10(sx + 1)
    scats_0, scats_1 = signal.reshape_features(sx, net)

    # show single window
    sx1 = np.squeeze(scats_1[channel]).T
    # cmap = nmmn.plots.wolframcmap()
    cmap = "Spectral_r"
    img = ax[0].pcolormesh(freq[0], freq[1], sx1[:-1, :-1], cmap=cmap)
    ax[0].set_xscale("log")
    ax[0].set_yscale("log")
    ax[0].set_xlim(freq[0].min(), freq[0].max())
    ax[0].set_ylim(freq[1].min(), freq[1].max())
    ax[0].set_xlabel("First-order frequency (Hz)")
    ax[0].set_ylabel("Second-order frequency (Hz)")
    ax[0].grid()
    ax[0].set_title("a", loc="left")

    cb = plt.colorbar(img, ax=ax[0], aspect=10)
    cb.set_label("Second-order\nscattering coefficients")

    # show first-oder scatterings
    ax[1].step(freq[0], scats_0[channel], where="post")
    ax[1].fill_between(freq[0], 0, scats_0[channel], alpha=0.2, step="post")
    ax[1].set_ylim(bottom=0, top=40)
    ax[1].set_xscale("log")
    ax[1].set_xlim(freq[0].min(), freq[0].max())
    ax[1].set_xlabel("First-order frequency (Hz)")
    ax[1].grid(which="both")
    ax[1].set_ylabel("First-order\nscattering coefficients")
    ax[1].set_title("b", loc="left")

    # show single window
    ax[2].plot(t, x[channel], "k")
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    ax[2].xaxis.set_major_locator(dateticks)
    ax[2].xaxis.set_major_formatter(datelabels)
    ax[2].set_xlim(t.min(), t.max())
    ax[2].set_ylim(-1, 1)
    ax[2].axvline(
        timestamp, c="C1", zorder=0, lw=4, alpha=0.4, label="Timestamp"
    )
    ax[2].legend()
    ax[2].grid()
    ax[2].set_ylabel("Amplitude")
    ax[2].set_title("c", loc="left")

    return fig


def show_scatterings(
    file_features,
    file_inventory,
    file_network,
    file_figure,
    timestamp,
    step,
    reader=None,
):
    """Show features on given time stamp."""
    # Read inventory
    tags, paths, start, end, _ = inventory.read(file_inventory)
    start, end = mdates.datestr2num(np.vstack((start, end)))

    # Get closest path to timestamp
    timestamp = mdates.date2num(timestamp)
    timestamp_index = np.searchsorted(start, timestamp) - 1
    timestamp_index = 0 if timestamp_index < 0 else timestamp_index
    path = paths[timestamp_index]
    tag = tags[timestamp_index]
    file_features = file_features.replace("*", tag)
    file_figure = file_figure.replace("*", tag)

    # load network
    net = pickle.load(open(file_network, "rb"))
    stdout("Loaded scatterings network from", file_network)
    bins = net.banks[0].bins

    # load features
    scatterings, scattering_times = load_feature_file(file_features)
    stdout("Loaded scatterings from", file_features)

    # Read seismograms
    stream = reader(path)
    stdout("Loaded seismogram from", path)
    times = stream[0].times("matplotlib")
    seismograms = [trace.data for trace in stream]
    npts = min((len(trace) for trace in seismograms))
    seismograms = np.array([component[:npts] for component in seismograms])
    times = times[:npts]
    step = int(step * stream[0].stats.sampling_rate)
    seismograms = signal.segmentize(seismograms, bins, step)
    times = signal.segmentize(times, bins, step)

    # Find corresponding window index
    window_index = np.searchsorted(scattering_times, timestamp) - 1
    window_index = 0 if window_index < 0 else window_index

    # select window
    scatterings = scatterings[window_index]
    t = times[window_index]
    x = seismograms[window_index]
    x = 0.9 * x / (np.abs(x).max() + 1e-5)

    # Reshape features
    sx = signal.reshape_features(scatterings, net)
    sx = signal.normalize_features(sx)
    sx = signal.vectorize_features(sx)

    # show
    show(t, x, sx, net, timestamp).savefig(file_figure)
    stdout("Saved figure at", file_figure + ".png")




# coding: utf-8
"""Common subcomands options for command-line interface.

Authors: Leonard Seydoux and Randall Balestriero
Email: leonard.seydoux@univ-grenoble-alpes.fr
Date: May, 2021
"""


import click
import os
import functools

COMMON_OPTIONS = [
    click.option(
        "--savepath",
        default=os.environ.get("SAVEPATH", "."),
        show_default=".",
        help="output directory",
        type=click.Path(resolve_path=True, writable=True),
    ),
    click.option(
        "--datapath",
        default=os.environ.get("DATAPATH", "."),
        show_default=".",
        help="input data path to scan",
    ),
    click.option(
        "--figpath",
        default=os.environ.get("FIGPATH", "."),
        show_default=".",
        help="path to save figures",
        type=click.Path(resolve_path=True, writable=True),
    ),
    click.option(
        "--start",
        type=click.DateTime(),
        default="0001-01-01 00:00:00",
        help="Select data after start time.",
    ),
    click.option(
        "--end",
        type=click.DateTime(),
        default="9999-01-01 00:00:00",
        help="Select data after start time.",
    ),
    click.option(
        "--channels",
        type=str,
        default="*",
        show_default=True,
        help="Channels to search for.",
    ),
    click.option(
        "--filename_inventory",
        type=str,
        default="inventory",
        show_default=True,
        help="Basename for inventory file.",
    ),
    click.option(
        "--filename_network",
        type=str,
        default="network",
        show_default=True,
        help="Basename for network file.",
    ),
    click.option(
        "--filename_reduction",
        type=str,
        default="reduction_{}",
        show_default=True,
        help="Basename for network file.",
    ),
    click.option(
        "--path_scatterings",
        type=str,
        default="scatterings",
        show_default=True,
        help="Name of scattering files directory.",
    ),
    click.option(
        "--path_features",
        type=str,
        default="features",
        show_default=True,
        help="Name of feature files directory.",
    ),
    click.option(
        "--path_clusters",
        type=str,
        default="clusters",
        show_default=True,
        help="Name of cluster files directory.",
    ),
    click.option(
        "--show",
        is_flag=True,
        default=False,
        show_default=True,
        help="Excluding flag to show or calculate.",
    ),
]


def common_options(f):
    return functools.reduce(lambda x, opt: opt(x), COMMON_OPTIONS, f)




# coding: utf-8
"""Show selected features in time and latent spaces."""

import nmmn.plots
import numpy as np
import os

from matplotlib import dates as mdates
from matplotlib import pyplot as plt
from scipy.stats import median_abs_deviation
from scipy.signal import medfilt

from ..io import stdout


def demad(x, factor=10.0):
    """Normalize signal with median absolute deviation.
    
    Arguments
    ---------
    x: np.ndarray
        The input signal.
    factor: float, optional
        An additional normalization factor.
    
    Returns
    -------
    The data normalized with median absolute deviation.
    """
    mad = median_abs_deviation(x)
    return x / np.mean(mad) / factor


def show_time(times, features, factor=0.4, medfilt_kernel=101):
    """Latent variables in time domain."""
    # Preprocess
    features = features.T
    features = demad(features)
    n_features, n_bins = features.shape

    # Figure
    fig, ax = plt.subplots(1, figsize=(5, 7))

    # Show
    for index, feature in enumerate(features):
        color = f"C{index % 3}"
        feature += index + 1
        feature_filtered = medfilt(feature, medfilt_kernel)
        ax.plot(times, feature, ".", ms=1, alpha=0.5, mew=0, color=color)
        ax.plot(times, feature_filtered, lw=0.7, color=color)

    # Labels
    ax.grid()
    ax.set_ylim(0, n_features + factor)
    ax.set_yticks(np.arange(n_features) + 1)
    ax.set_ylabel("Independant component index")

    # Date labels
    dateticks = mdates.AutoDateLocator()
    datelabels = mdates.ConciseDateFormatter(dateticks)
    ax.xaxis.set_major_locator(dateticks)
    ax.xaxis.set_major_formatter(datelabels)
    ax.set_xlim(times.min(), times.max())

    # Remove borders
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.tick_params(axis="y", length=0)

    return fig


def show_latent(features, cmap=nmmn.plots.wolframcmap(), nbins=800):
    """Latent variables in versus diagram."""
    # Preprocess
    features = features.T
    features = demad(features)
    n_features = features.shape[0] - 1

    # Figure
    figsize = 2 * [n_features]
    gridspec_kw = dict(hspace=0.1, wspace=0.1)
    fig, ax = plt.subplots(
        n_features,
        n_features,
        figsize=figsize,
        gridspec_kw=gridspec_kw,
        constrained_layout=False,
        sharex="col",
        sharey="row",
    )

    # Versus diagrams
    for i in range(n_features):

        x = features[i]
        x_min, x_max = x.min(), x.max()
        x_bins = np.linspace(x_min, x_max, nbins)

        for j in range(0, n_features):

            y = features[j + 1]
            y_min, y_max = y.min(), y.max()
            y_bins = np.linspace(y_min, y_max, nbins)

            # Lower triangular
            if j >= i:

                # Histogram
                counts, _, _ = np.histogram2d(x, y, (x_bins, y_bins))
                counts = counts.T
                counts[counts == 0] = 1e-4
                counts = np.log(counts)
                extent = [x_min, x_max, y_min, y_max]
                ax[j, i].imshow(
                    counts, cmap=cmap, extent=extent, aspect="auto"
                )
                ax[j, i].grid()
                ax[j, i].set_xticks([])
                ax[j, i].set_yticks([])
                ax[j, i].set_ylim(y_min, y_max)
                ax[j, i].set_xlim(x_min, x_max)

                # Style
                for side in ax[j, i].spines:
                    ax[j, i].spines[side].set_visible(False)

            # Upper triangular
            else:
                ax[j, i].set_axis_off()

            # Labels
            if j == n_features - 1:
                ax[j, i].set_xlabel(f"Latent {i + 1}")
            if i == 0:
                ax[j, i].set_ylabel(f"Latent {j + 2}")

    return fig


def show_features(file_features, file_figure, medfilt_kernel=101):

    # Basename
    basename = os.path.basename(file_features)
    basename = basename.split(".")[0]

    # Load features
    with np.load(file_features) as data:
        features = data["features"]
        times = data["times"]
    stdout("Loaded features from", file_features)

    # Show in time and save
    fig = show_time(times, features, medfilt_kernel=medfilt_kernel)
    print("figure created")
    file_figure_time = file_figure + "_time.png"
    fig.show()
    fig.savefig(file_figure_time, dpi=600)
    print("figure saved")
    stdout("Figure saved at", file_figure_time)

    # Show in latent space and save
    fig = show_latent(features)
    file_figure_space = file_figure + "_space.png"
    fig.savefig(file_figure_space)
    stdout("Figure saved at", file_figure_space)




# coding: utf-8
"""Routines for inventorizing data.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""

import obspy
import pandas as pd

from glob import glob
from parse import parse
from tqdm import tqdm

from .io import stdout


def inventorize(parsable_path, channel_pattern, tag_pattern):
    """Glob files with expansion on tag and channel, and get headers.
    
    Arguments
    ---------
    parsable_path: str
        The path to data with a tag and a channel string variables.
    channel_pattern: str
        Pattern to expand for channels.
    tag_pattern: str
        Pattern to expand for tag.
    
    Returns
    -------
    db: pandas.DataFrame
        The extracted availability and metadata.
    """
    # File pattern creation
    pattern = dict(tag=tag_pattern, channel=channel_pattern)
    filepath_pattern = parsable_path.format(**pattern)
    stdout("Data files pattern {}", filepath_pattern)

    # File pattern glob
    filepaths_matching = sorted(glob(filepath_pattern))
    if len(filepaths_matching) > 0:
        stdout("Found {} matching files", len(filepaths_matching))
    else:
        print("No files matching pattern; exiting.")
        exit()

    # Build up inventory
    header = obspy.read(filepaths_matching[0], headonly=True)[0].stats
    db_entrynames = [item for item in header]
    db = pd.DataFrame(columns=db_entrynames)

    # Extract headers from every files
    index = 0
    for filepath in tqdm(filepaths_matching, desc="Making inventory"):

        # Fix Windows filepath (temporary solution)
        if  "\\" in filepath:
            filepath=filepath.replace("\\", "/")

        # Read headers
        stream = obspy.read(filepath, headonly=True)
        parsed = parse(parsable_path, filepath)
        # print(parsable_path)
        # print(filepath)

        # Extract stats from every traces
        for trace in stream:
            for entry in db_entrynames:
                db.loc[index, entry] = trace.stats[entry]
            db.loc[index, "tag"] = parsed["tag"]
            db.loc[index, "path"] = filepath
            index += 1

    return db


def read(filename_inventory):
    """Read inventory pickle file.
    
    Arguments
    ---------
    filename: str
        The filename of the inventory.

    Returns
    -------
    pandas.DataFrame
        The tags and corresponding metadata to read.
    """
    # Read pickle file
    db = pd.read_pickle(filename_inventory)

    # Drop columns
    db = db.drop(db._format.unique()[0].lower(), axis=1)
    db = db.drop("_format", axis=1)

    # Convert to pandas timestamps
    db.starttime = pd.to_datetime([t.datetime for t in db.starttime])
    db.endtime = pd.to_datetime([t.datetime for t in db.endtime])

    # Infer types
    db = db.infer_objects()

    # Duration
    db["duration"] = db.endtime - db.starttime
    db["duration_hours"] = db.duration.dt.total_seconds() / 3600

    return db



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Clustering.

author:
    Leonard Seydoux and Randall Balestriero
"""

import numpy as np

from fastcluster import linkage_vector
from scipy.cluster import hierarchy


def get_leaves(d, ax):
    """Get dendrogram list of leaves coordinates and colors.

    Arguments
    ---------
    d: dict
        Output of the scipy.hierarchy.dendrogram function.

    Returns
    -------
    coords, colors: array-like
        The x coordinates and color of each leave.
    """
    coords = list()
    for icoord, dcoord in zip(d["icoord"], d["dcoord"]):
        if dcoord[0] == 0:
            coords.append(icoord[0])

        if dcoord[-1] == 0:
            coords.append(icoord[-1])

    isort = np.argsort(coords)
    coords = np.array(coords)[isort]

    # Cardinality
    card = list()
    for y in ax.get_yticklabels():
        y = y.get_text()
        card.append(int(y[1:-1]) if "(" in y else 1)
    card = np.array(card)
    return coords, card


def linkage(x, distance="ward"):
    """Calculate linkage matrix on reduced matrix.

    The reduction is performed with independant component analysis.

    Arguments
    ---------
    x: np.ndarray
        Design matrix.

    Keyword arguments
    -----------------
    reduction: int
        The number of dimensions to consider for the linkage calculation.

    distance: str
        The metric to use for calculating the linkage matrix.

    Returns
    -------
    link:
        Linkage matrix.
    """
    # linkage
    link = linkage_vector(x, distance)
    print("Linkage calculated")
    return link


def fcluster(link, threshold):
    """Cluster data based on linkage matrix and threshold.

    Arguments
    ---------
    linkage:
        The linkage matrix

    threshold:
        The threshold to cluster the data with from a distance argument.

    Returns
    -------
    clusters: np.ndarray
        The per-sample cluster attribution.
    """
    return hierarchy.fcluster(link, t=threshold, criterion="distance")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Display routines."""

import numpy as np
import os
import seaborn as sns

from . import cluster

from matplotlib import pyplot as plt
from matplotlib import dates as mdates


def set_style(name):
    """Set matplotlib style for the current session.

    This function affects the display codes globally. For a more specific style
    manipulation, the plt.rc_context is more adapted.

    Arguments
    ---------
    name: str
        The style name. This name should be registered in the config repository
        of Matplotlib.

    Returns
    -------
    The returns of the plt.style.use function.
    """
    return plt.style.use(name)


def dateticks(axis):
    """Automatic date formatting for given axes.

    This function wraps the AutoDateLocator and the ConciseDateFormatter
    routines of matplotlib for better code readability.

    Arguments
    ---------
    axis: plt.Axis()
        The axis instance to format with automatic dateticks. This can be for
        instance the currrent x-axis for the current axes (ax.xaxis).
    """
    locator = mdates.AutoDateLocator()
    axis.set_major_locator(locator)
    axis.set_major_formatter(mdates.ConciseDateFormatter(locator))
    pass


def symmetrize(axis):
    """Force symmetric limits for the given axis.

    Arguments
    ---------
    axis: plt.Axis()
        The axis instance to format with automatic dateticks. This can be for
        instance the currrent x-axis for the current axes (ax.xaxis).
    """
    limits = axis.get_view_interval()
    maximum = np.abs(limits).max()
    axis.set_view_interval(-maximum, maximum)
    pass


def save(figure, directory, filename):
    """Save given axes figure in output directory.

    This function behaves similarly than the scat.io.save function for pickle
    files but for images. The output image format can be specified with adding
    an explicit extension.

    Note that the figure is closed at the end of the routine in order to
    prevent large number of opened figures in loops.

    Arguments
    ---------
    figure: plt.Figure()
        The current figure instance.

    directory: str
        The current output directory.

    filename: str
        The figure file name.
    """
    # figure name and save
    filename = os.path.join(directory, f"{filename}")
    figure.savefig(filename)

    # checkout
    # logger.info(f"saved figure at {bold(filename)}")

    # close figure to disable large number of figures in loops
    plt.close(figure)


def show_dendrogram(linkage, threshold, ax=plt.gca(), depth=30):
    """Show dendrogram and returns basic cluster informations.

    Arguments
    ---------
    linkage: np.ndarray
        Output of the linkage function.

    threshold: float
        Threshold value from which the clustering is made.

    Keyword arguments
    -----------------
    ax: plt.Axes
        The axes to draw the dendrogram into.

    depth: str
        The dendrogram depth

    Return
    ------

    """
    # Show and get dendrogram
    with plt.rc_context({"lines.linewidth": 0.7}):
        dendrogram_infos = cluster.hierarchy.dendrogram(
            linkage,
            p=depth,
            truncate_mode="lastp",
            color_threshold=0,
            ax=ax,
            orientation="left",
            above_threshold_color="0.3",
            count_sort=True,
        )

    # Extract informations
    coords, sizes = cluster.get_leaves(dendrogram_infos, ax)
    colors = sns.color_palette("husl", n_colors=len(coords))

    # Get leave indexes
    indexes_flat = cluster.hierarchy.leaves_list(linkage)
    indexes = list()
    i = 0
    for size in sizes:
        indexes.append(indexes_flat[i : i + size])
        i += size

    # Labels
    ax.set_ylim(bottom=2.5)
    ax.set_xlabel("Rescaled distance")
    ax.grid(axis="x")
    ax.tick_params(width=0)
    for side in ax.spines:
        ax.spines[side].set_visible(0)

    # Leaves
    for coord, color in zip(coords, colors):
        ax.plot(0, coord, ".", ms=10, c=color, clip_on=False)

        index = int((coord - 5) / 10) + 1
        label = "{:d} ".format(index)
        ax.text(-0.1, coord, label, color=color, va="center", weight="bold")

    return coords, sizes, colors, indexes


def violin(data, bins, bottom=0, ax=plt.gca(), **kwargs):
    """Handmade violin from fill_between.

    Arguments
    ---------
    x, y: np.ndarray
        The data to violin.

    Keyword arguments
    -----------------
    bottom: float
        The baseline.

    ax: plt.Axes
        The axes to draw into.
    """
    y, x = np.histogram(data, bins)
    # bottom -= 2.5
    if len(y):
        y = 5.0 * y / y.max()
        x = x[:-1]
        kwargs.setdefault("clip_on", False)
        ax.step(x, bottom + y, lw=0.2, where="post", **kwargs)
        ax.fill_between(x, bottom, bottom + y, alpha=1, step="post", **kwargs)


def show_cluster(link, threshold, times, depth=30):
    """Display the main clustering summary.

    Arguments
    ---------
    link:
        The linkage matrix.

    clusters: np.ndarray
        The per-samples cluster attribution.

    times: np.ndarray
        The time indexes

    Returns
    -------
    fig, ax: plt.Figure, plt.Axes
        The matplotlib figure and axes instance.
    """
    # deactivate axes basic properties
    spines_off = {
        "axes.spines.right": False,
        "axes.spines.bottom": False,
        "axes.spines.top": False,
        "axes.facecolor": "none",
        "xtick.top": False,
        "xtick.bottom": False,
    }

    # generate axes
    gs = dict(width_ratios=[2, 3, 1, 2])
    figsize = 7, depth * 0.35
    with plt.rc_context(spines_off):
        fig, ax = plt.subplots(
            1, 4, sharey=True, figsize=figsize, gridspec_kw=gs
        )

    # occurences vectors
    dt = times[1] - times[0]
    calendar = np.linspace(times[0], times[-1], 100)
    daily = np.linspace(0, 24, 24)

    # show dendrogram
    leaves = show_dendrogram(link, threshold, ax=ax[0], depth=depth)

    # loop over clusters
    for clust, size, color, indexes in zip(*leaves):

        # overall duration
        ratio = 100 * size / len(times)
        duration = f" {size}"
        ax[-1].barh(clust, ratio, height=7, color=color)
        ax[-1].text(ratio, clust, duration, color=color, size=6, va="center")

        # occurences
        if len(indexes) > 1:

            # calendar occurence
            violin(times[indexes], calendar, clust, ax[1], color=color)

            # daily occurence
            hours = 24 * (times[indexes] - np.fix(times[indexes]))
            violin(hours, daily, clust, ax=ax[2], color=color)

    # labels
    ax[-1].set_xlabel("Relative size (%)")
    ax[1].set_xlabel("Calendar occurence")
    ax[1].set_ylabel("Normalized detection rate")
    # ax[1].set_yticks(range(len(l)))
    ax[2].set_xlabel("Local time")
    # ax[4].set_xlabel("Frequency (Hz)")
    ax[0].set_title("a", loc="left", weight="bold")
    ax[-1].set_title("d", loc="left", weight="bold")
    ax[1].set_title("b", loc="left", weight="bold")
    ax[2].set_title("c", loc="left", weight="bold")
    # ax[4].set_title("e\n", loc="left", weight="bold")

    # calendar occurence
    locator = mdates.AutoDateLocator()
    formatter = mdates.ConciseDateFormatter(locator, show_offset=False)
    ax[1].set_xlim(times.min(), times.max())
    ax[1].xaxis.set_major_locator(locator)
    ax[1].xaxis.set_major_formatter(formatter)
    plt.setp(ax[1].get_xticklabels(), rotation="vertical")

    # daily occurence
    ax[2].set_xlim([0, 24])
    ax[2].set_xticks([0, 6, 12, 18, 24])
    ax[2].set_xticklabels(["00:00", "06:00", "12:00", "18:00", "00:00"])
    plt.setp(ax[2].get_xticklabels(), rotation="vertical")

    # scatterings
    # ax[4].set_xlim(frequencies.min(), frequencies.max())
    # ax[4].set_xscale("log")
    # ax[4].set_xticks([1, 10])

    # grids
    for a in ax[1:]:
        a.grid(axis="x")
    for a in ax:
        a.set_yticklabels([])

    # return
    return fig, ax, leaves


def show_cumulated(times, leaves, clusters):
    """Show cumulated detection curves."""
    fig, ax = plt.subplots(1, figsize=(8, 8))
    for clust, size, color, indexes in zip(*leaves):
        indexes = sorted(indexes)
        cumulated = np.cumsum(np.ones_like(times[indexes]))
        ax.step(times[indexes], cumulated, color=color)

    ax.grid(lw=0.2)
    ax.set_xlim(times.min(), times.max())
    dateticks = mdates.AutoDateLocator()
    ax.xaxis.set_major_locator(dateticks)
    ax.xaxis.set_major_formatter(mdates.ConciseDateFormatter(dateticks))
    ax.set_ylabel("Cumulated detections")
    return fig, ax



# coding: utf-8
"""Command-line module for data inventory.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""

import click
import os

from .common import common_options

from ..inventory import inventorize
from .. import io


@click.command("inventory", short_help="Create dataset inventory.")
@common_options
@click.option(
    "--tags",
    type=str,
    default="*",
    show_default=True,
    help="Tags to search for in the datapath. Accepts wildcards.",
)
def inventory(
    tags=None,
    channels=None,
    savepath=None,
    datapath=None,
    filename_inventory=None,
    **kwargs,
):
    """Create an inventory of available data.

    The script generates an inventory of the data files from a parsable path.
    This inventory allows later to connect the calculated features and the
    input data quickly. In addition, this script enables selecting the input
    data based on different criteria, such as the sampling rate, duration,
    channel, or dates. 

    The parser determines two elements from the given parsable path: a "tag"
    identifier for each file and a "channel" variable. Note that these two
    variables can appear multiple times in the string. 

    Let us consider the example list of files below containing three days of
    data (tagged as 2014.14, 2014.15, and 2014.16) recorded with two channels
    (E and Z).

        /path/to/data/2010.014/HHE/2010.014.HHE.sac
        /path/to/data/2010.014/HHZ/2010.014.HHZ.sac
        /path/to/data/2010.015/HHE/2010.015.HHE.sac
        /path/to/data/2010.015/HHZ/2010.015.HHZ.sac
        /path/to/data/2010.016/HHE/2010.016.HHE.sac
        /path/to/data/2010.016/HHZ/2010.016.HHZ.sac

    The idea is to generate a list of file paths that can read all channels of
    the same date at once with the obspy's read routine with the following
    command: 

        scatseisnet inventory --datapath /path/to/data/{tag}/HH{channel}/{tag}.HH{channel} --channels Z E --tag * --filename_inventory inventory

    From this command, the parser will determine the following list of tags
    (that can be restricted to more specific regular expression with the --tag
    command-line argument) with corresponding paths and save them in the
    "inventory" file specified in the --file_inventory option.

        2014.014 /path/to/data/2010.014/HH[Z,E]/2010.014.HH[Z,E].sac
        2014.015 /path/to/data/2010.015/HH[Z,E]/2010.015.HH[Z,E].sac
        2014.016 /path/to/data/2010.015/HH[Z,E]/2010.016.HH[Z,E].sac

    Note that the notation "[Z,E]" enables obspy to read both channels Z and E
    in the same stream at once, given the path expansion capabilities of the
    obspy read routine (based on the glob Python library). 
    """
    # Resolve paths
    dirpath_inventory = os.path.join(savepath, "inventories")
    filepath_inventory = os.path.join(dirpath_inventory, filename_inventory)

    # Calculate inventory
    io.mkdir(dirpath_inventory)
    db = inventorize(datapath, channels, tags)

    # Save inventory
    db.to_pickle(filepath_inventory)
    io.stdout("Inventory complete saved at", filepath_inventory)




# coding: utf-8
"""Command-line interface module.

The subcommands are defined in dedicated modules. Note that all subcommands
share a subset of options defined in the "common.py" module. The "main"
function defined here is a group for attaching all the subcommands, and it
establishes the program's entry point.

Author: Leonard Seydoux Date: June, 2021 Email:
leonard.seydoux@univ-grenoble-alpes.fr
"""

import click

from .inventory import inventory
from .linkage import linkage
from .features import features
from .transform import transform
from .waveforms import waveforms


@click.group()
def main():
    """Deep scattering transform on segmented time series.

    This program contains a series of command-line tools for clustering
    continuous time series with a deep scattering network. The following
    sub-commands must be run in a specific order from the continuous data to
    the cluster results.

    1. The inventorize command lists and selects the data based on usual meta
    parameters (sampling rate, duration, channels) and stores the relevant
    datapaths into an inventory file. This first command helps explore the data
    coverage in time, selecting appropriate time segments, and running the
    remaining steps on the actual data.

    2. The transform command runs the deep scattering transform on the
    segmented time series and stores the scattering coefficients for later
    feature extraction. 

    3. With the features command, the large-dimensional scattering coefficients
    are reduced to a low-dimensional space which dimensions are considered
    features here. These features are used in the clustering step next.

    4. The command calculates the linkage matrix that helps cluster the data
    based on some criteria of similarity (a metric and a method). Once this
    matrix is calculated, the clusters are extracted for further analyses.

    Created in May, 2021 by Leonard Seydoux
    (leonard.seydoux@univ-grenoble-alpes.fr) and Randall Balestriero.

    """
    pass


# Add subcommands
main.add_command(inventory)
main.add_command(transform)
main.add_command(features)
main.add_command(linkage)
main.add_command(waveforms)



# coding: utf-8
"""Command-line module for data inventory.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""

import click
import numpy as np
import os
import pickle

from sklearn.decomposition import FastICA as skmodel
from sklearn.metrics import r2_score, mean_squared_error
from .common import common_options

from .. import signal, io
from ..display import show_features


@click.command("features", short_help="Calculate features.")
@common_options
@click.option(
    "--dimensions",
    default=10,
    show_default=True,
    help="Number of latent space dimensions.",
    type=int,
)
@click.option(
    "--normalize",
    is_flag=True,
    default=False,
    show_default=True,
    help="Normalization flag.",
)
@click.option(
    "--medfilt",
    type=int,
    default=101,
    show_default=True,
    help="Median filter kernel size.",
)
def features(
    dimensions,
    normalize=False,
    savepath=None,
    figpath=None,
    filename_network=None,
    filename_reduction=None,
    path_scatterings=None,
    path_features=None,
    show=False,
    medfilt=None,
    **kwargs,
):
    """Reduce scattering domain dimensions."""
    # Path
    dirpath_models = os.path.join(savepath, "models")
    dirpath_scats = os.path.join(savepath, path_scatterings)
    dirpath_features = os.path.join(savepath, path_features)
    dirpath_figure = os.path.join(figpath, "features")

    # Files
    filepath_network = os.path.join(dirpath_models, filename_network)
    filepath_reduction = os.path.join(dirpath_models, filename_reduction)
    filepath_scatterings = os.path.join(dirpath_scats, "scatterings_*.npz")
    filepath_features = os.path.join(dirpath_features, "features_{}_{}.npz")
    filepath_figure = os.path.join(dirpath_figure, "features_{}_{}")

    # Append parameters to filenames
    norm = "norm" if normalize is True else "no-norm"
    filepath_reduction = filepath_reduction.format(dimensions)
    filepath_features = filepath_features.format(dimensions, norm)
    filepath_figure = filepath_figure.format(dimensions, norm)

    if show is True:
        io.mkdir(dirpath_figure)
        show_features(
            filepath_features, filepath_figure, medfilt_kernel=medfilt
        )
    else:
        # Directories
        io.mkdir(dirpath_features)

        # Parameters
        io.stdout("Using {} for scattering coefficients", filepath_scatterings)
        io.stdout("Using {} for features", filepath_features)
        io.stdout("Using {} dimensions", dimensions)
        io.stdout("Using normalization", normalize)

        # Load features
        features, times = io.load_features(filepath_scatterings)

        # Normalize features
        if normalize is True:
            io.stdout("Read newtork from", filepath_network)
            net = pickle.load(open(filepath_network, "rb"))
            for index in range(features.shape[0]):
                feature = signal.reshape_features(features[index], net)
                feature = signal.normalize_features(feature)
                features[index] = signal.vectorize_features(feature)

        # Preprocess
        keep = features.sum(axis=1) > 1e-3
        times = times[keep]
        features = features[keep]
        features = np.log10(features + 1e-3)
        features = (features - features.min()) / (
            features.max() - features.min()
        )

        # Reduce
        print("Performing reduction")
        model = skmodel(n_components=dimensions)
        latents = model.fit_transform(features)
        inversed = model.inverse_transform(latents)
        io.stdout("Mean squared error", mean_squared_error(features, inversed))
        io.stdout("R2 coefficient", r2_score(features, inversed))

        # Save latent variables
        np.savez(filepath_features, times=times, features=latents)
        io.stdout("Saved features at", filepath_features)

        # save model
        pickle.dump(model, open(filepath_reduction, "wb"))
        io.stdout("Saved model at", filepath_reduction)

    pass



# coding: utf-8
"""Command-line module for data inventory.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""

import click
import numpy as np
import obspy
import os
import pandas as pd
import pickle

from matplotlib import dates as mdates
from obspy import read, Stream, UTCDateTime
from tqdm import tqdm

from .common import common_options

from .. import inventory, signal, Network, io
from ..display import show_scatterings


# def load_waveform(path, trim=None):
#     """Waveform reader."""
#     stream = read(path)
#     if trim is not None:
#         stream.trim(*trim)
#     stream.merge(method=1, fill_value="interpolate")
#     stream.detrend()
#     stream.decimate(5)
#     stream.filter(type="highpass", freq=5.0)
#     stream.taper(max_percentage=0.05)
#     return stream


def load_waveform(paths):

    stream = obspy.Stream()
    for path in paths:
        stream += obspy.read(path)
    stream.detrend("linear")
    stream.merge(method=1)
    stream.detrend("linear")
    stream.decimate(4)
    stream.detrend("linear")
    stream.filter(type="highpass", freq=1)
    starttime = max([tr.stats.starttime for tr in stream])
    endtime = min([tr.stats.endtime for tr in stream])
    stream.trim(starttime, endtime, fill_value=0)
    return stream


@click.command("transform", short_help="Calculate scattering coefficients.")
@common_options
@click.option(
    "--segment",
    type=float,
    default=200.0,
    show_default=True,
    help="Segment duration (seconds).",
)
@click.option(
    "--step",
    type=float,
    default=100.0,
    show_default=True,
    help="Step duration (seconds).",
)
@click.option(
    "--pooling",
    "-p",
    type=click.Choice(("max", "avg", "med")),
    default="max",
    show_default=True,
    help="Pooling type.",
)
@click.option(
    "--octaves",
    "-j",
    multiple=True,
    default=[4, 10],
    show_default=True,
    help="Number of octaves per layer.",
)
@click.option(
    "--resolution",
    "-q",
    multiple=True,
    default=[2, 2],
    show_default=True,
    help="Number of wavelets per octaves per layer.",
)
@click.option(
    "--quality",
    "-d",
    multiple=True,
    default=[2, 2],
    show_default=True,
    help="Quality factor per layer.",
)
@click.option(
    "--timestamp",
    type=click.DateTime(),
    default=None,
    help="Dislay given timestamp (show must be on).",
)
@click.option(
    "--reader",
    type=str,
    default="reader.py",
    help="Module where the load function is defined.",
)
def transform(
    segment,
    step,
    pooling,
    octaves,
    resolution,
    quality,
    savepath=None,
    filename_inventory=None,
    filename_network=None,
    path_scatterings=None,
    start=None,
    end=None,
    reader=None,
    **kwargs,
):
    """Transform seismograms into scattering domain."""
    # Directories
    dirpath_inventory = os.path.join(savepath, "inventories")
    dirpath_models = os.path.join(savepath, "models")
    dirpath_scats = os.path.join(savepath, path_scatterings)

    # Files
    filepath_network = os.path.join(dirpath_models, filename_network)
    filepath_inventory = os.path.join(dirpath_inventory, filename_inventory)
    filepath_scatterings = os.path.join(dirpath_scats, "scatterings_*.npz")

    # Directories
    io.mkdirs(dirpath_models, dirpath_scats)

    # Parameters
    io.stdout("Using {} for scattering coefficients", filepath_scatterings)
    io.stdout("Using {} sec windows with {} sec step", (segment, step))
    io.stdout("Using {} octaves", list(octaves))
    io.stdout("Using {} resolutions", list(resolution))
    io.stdout("Using {} quality factors", list(quality))
    io.stdout("Using {} pooling", pooling)

    # Read inventory
    database = inventory.read(filepath_inventory)

    # Trim database
    start = pd.to_datetime(start, errors="coerce")
    end = pd.to_datetime(end, errors="coerce")
    if not pd.isna(start):
        database = database[database.starttime > start]
    if not pd.isna(end):
        database = database[database.starttime < end]

    # Initialize progress bar with reading waveforms list
    progress_bar = tqdm(database.groupby("tag"))

    # Transform loop
    for tag, db in progress_bar:

        # Extract streams
        paths = db.path
        stream = load_waveform(paths)

        # Init nertwork
        if 0 in db.index:
            sampling_rate = stream[0].stats.sampling_rate
            bins = int(segment * sampling_rate)
            net = Network(octaves, resolution, quality, bins, sampling_rate)
            pickle.dump(net, open(filepath_network, "wb"))
            io.stdout("Saved network at", filepath_network)

        # Update progress bar description with current file
        progress_bar.set_description(f"Process {tag}")

        # Turn into array
        npts = np.unique([len(tr) for tr in stream])
        if len(npts) > 1:
            continue
        times = stream[0].times("matplotlib")
        seismograms = np.array([trace.data for trace in stream])
        # npts = min((len(trace) for trace in seismograms))
        # seismograms = np.array([channel[:npts] for channel in seismograms])
        # times = times[:npts]

        # Extract segments
        if npts < bins:
            continue
        slide = int(step * sampling_rate)
        seismograms = signal.segmentize(seismograms, bins, slide)
        times = signal.segmentize(times, bins, slide)[:, 0]

        # Scattering Network
        features = net.transform(seismograms, pooling)

        # Save features
        file_features = filepath_scatterings.replace("*", tag)
        features = {f"features_{i}": f for i, f in enumerate(features)}
        np.savez(file_features, times=times, **features)

    pass


def get_all_subkeys(dictionnary, subkey):
    """Get all sub entries of a dictionnary with given subkey."""
    subvalues = [dictionnary[key][subkey] for key in dictionnary.keys()]
    return subvalues



# coding: utf-8
"""Display features.

Author: Leonard Seydoux
Date: June, 2021
Email: leonard.seydoux@univ-grenoble-alpes.fr
"""

import numpy as np
import obspy

from matplotlib import dates as mdates
from matplotlib import pyplot as plt
from scipy.spatial.distance import euclidean

# from scipy.stats import median_abs_deviation as mad

from .. import inventory


plt.rcParams["figure.constrained_layout.use"] = True
COLORS = "#226fc7", "#22c78b", "#c75c22"


def match_tag(timestamp, times):
    try:
        return np.where((np.fix(timestamp) - times) == 0)[0][0]
    except:
        return np.argmin(np.abs(np.fix(timestamp) - times))


def show_waveforms(
    segment,
    file_features,
    file_prediction,
    file_inventory,
    file_figure,
    n_samples=None,
    reader=None,
    factor=0.8,
):

    # Features
    with np.load(file_features) as data:
        feature_times = data["times"]
        feature_values = data["features"]

    # Predictions
    with np.load(file_prediction) as data:
        predictions = data["predictions"]

    # Get start times of inventory
    database = inventory.read(file_inventory)
    # tags = list(database.keys())
    # inventory.read(file_inventory)
    # _, paths, starts, *_ = inventory.read(file_inventory)
    # starts = [mdates.datestr2num(t) for t in starts]
    # starts = get_all_subkeys(database, "date_start") #does not work
    starts = [database["starttime"]]
    starts = mdates.date2num(starts)
    # paths = get_all_subkeys(database, "path") #does not work
    paths = [database["path"]]
    # end = get_all_subkeys(database, "date_end")
    # zeros = get_all_subkeys(database, "n_zeros")
    # sampling_rate = get_all_subkeys(database, "sampling_rate")
    # npts = get_all_subkeys(database, "n_pts")
    # start = mdates.date2num(start)
    # end = mdates.date2num(end)

    # Calculate centroids
    classes = np.unique(predictions)

    for cluster in classes:

        # Extract clusters
        within_cluster = predictions == cluster
        cluster_samples = feature_values[within_cluster]
        cluster_times = feature_times[within_cluster]

        # Centroid
        centroid = np.mean(cluster_samples, axis=0)
        distances = list()
        for sample in cluster_samples:
            distances.append(euclidean(sample, centroid))
        distances = np.array(distances)

        # Extract best waveforms timestamps
        distances_argsort = np.argsort(distances)
        sorted_times = cluster_times[distances_argsort][:n_samples]
        # sorted_times = cluster_times[:n_samples]

        # Create figure
        fig, axes = plt.subplots(
            nrows=1,
            ncols=2,
            figsize=(6, 6),
            sharey=True,
            gridspec_kw={"width_ratios": (2, 1)},
        )

        y_ticklabels = list()
        for event_id, timestamp in enumerate(sorted_times):

            # Match time
            print(timestamp)
            tag_id = match_tag(timestamp, starts)
            start = obspy.UTCDateTime(mdates.num2date(timestamp))
            end = obspy.UTCDateTime(
                mdates.num2date(timestamp + segment / 3600 / 24)
            )
            # stream = reader(paths[tag_id], trim=(start, end)) #ERROR trim option not accepted
            stream = reader(paths[tag_id])
            n_channels = len(stream)
            if n_channels == 0:
                continue

            y_ticklabels.append(
                mdates.num2date(timestamp).strftime("%y/%m/%d %H:%M:%S")
            )

            # Show
            plot_style = dict()
            # components_max = np.mean([mad(trace.data) for trace in stream])
            components_max = np.max([np.max(trace.data) for trace in stream])
            for trace_id, trace in enumerate(stream):

                # Extract data
                trace_times = trace.times()
                trace_data = trace.data
                # trace_data = trace_data / components_max / 20
                trace_data = trace_data / components_max
                trace_data += factor * trace_id + n_channels * event_id

                # Assign channel color and label
                plot_style["color"] = COLORS[trace_id]
                if event_id == 0:
                    plot_style["label"] = trace.stats.channel

                # Plot
                axes[0].plot(trace_times, trace_data, **plot_style)

            axes[0].plot(
                [0, 0],
                [
                    event_id * n_channels,
                    event_id * n_channels + factor * (n_channels - 1),
                ],
                "0.3",
                clip_on=False,
            )

            axes[0].plot(
                2 * [trace_times.max()],
                [
                    event_id * n_channels,
                    event_id * n_channels + factor * (n_channels - 1),
                ],
                "0.3",
                clip_on=False,
            )

        # Skip if no waveforms detected
        if not y_ticklabels:
            continue

        # Yticks
        y_ticks = np.arange(0, n_channels * len(y_ticklabels), n_channels)
        y_ticks = y_ticks + factor

        # Distances
        axes[1].barh(
            y_ticks,
            distances[distances_argsort][:n_samples],
            height=1.5,
            color="0.8",
        )
        axes[1].axvline(np.mean(distances))
        # axes[1].plot(distances[distances_argsort][:n_samples], y_ticks, "o")
        axes[1].set_xlabel("Euclidean distance (latent space)")
        axes[1].spines["top"].set_visible(False)
        axes[1].spines["right"].set_visible(False)

        # Labels
        axes[0].legend(loc="upper center", ncol=3)
        axes[0].set_xlim(0, trace_times.max())
        axes[0].set_xlabel("Time (seconds)")
        axes[0].set_yticks(y_ticks)
        axes[0].set_ylim(-1, n_channels * n_samples + 1)
        axes[0].set_yticklabels(y_ticklabels, size="small")
        axes[0].grid(clip_on=False)
        axes[0].set_title(f"Cluster {cluster}", loc="left")
        axes[0].spines["top"].set_visible(False)
        axes[0].spines["right"].set_visible(False)
        axes[0].spines["left"].set_visible(False)
        axes[0].tick_params(axis="y", length=0, which="both")

        # # Latent inset
        # inset = fig.add_axes([0.8, 0.8, 0.2, 0.2])
        # inset.plot(feature_values[:, 0], feature_values[:, 1], ".", c="0.6")
        # inset.plot(
        #     cluster_samples[:, 0], cluster_samples[:, 1], ".", c="k",
        # )
        # inset.plot(
        #     cluster_samples[distances_argsort, 0][:n_samples],
        #     cluster_samples[distances_argsort, 1][:n_samples],
        #     ".",
        #     c="C1",
        # )

        fig.savefig(file_figure.replace("*", f"{cluster:02d}"))
        plt.close()


def get_all_subkeys(dictionnary, subkey):
    """Get all sub entries of a dictionnary with given subkey."""
    subvalues = [dictionnary[key][subkey] for key in dictionnary.keys()]
    return subvalues



# -*- coding: utf-8 -*-
"""Scattering network.

author:
    Leonard Seydoux and Randall Balestriero
"""

import numpy as cp

from .signal import pool
from .wavelet import ComplexMorletBank


class Network:
    """Scattering network."""

    def __init__(self, octaves, resolution, quality, bins, sampling_rate=1.0):
        """Initialize graph.

        Arguments
        ---------
        args: :class:`argparse.Namespace`
            The input arguments. Since most of the input arguments are
            dedicated to the scattering newtork architecture, they are all
            passed to this constructor. Note that almost all the attributes
            of the args are attributed to the present class.

        Keyword arguments
        -----------------
        sampling: float (optional)
            The input data sampling rate. This is useful to keep track of
            physical frequencies in the filterbanks properties.

        channels: list (optional)
            A list with channels names. At the time this is used only for
            getting the number of channels for each data samples, but it can
            later be used for naming the scalograms and scatterings.
        """

        # define banks
        bank_properties = zip(octaves, resolution, quality)
        self.banks = [ComplexMorletBank(bins, *p) for p in bank_properties]
        self.sampling_rate = sampling_rate
        pass

    def transform_sample(self, sample, reduce_type=None):
        """Scattering network transformation.

        Arguments
        ---------
        sample: np.ndarray
            The input 
        """
        # loop over banks
        input_sample = sample
        output = list()
        for bank in self.banks:
            scalogram = bank.transform(input_sample)
            input_sample = scalogram
            output.append(pool(scalogram, reduce_type))
        return output

    def transform(self, samples, reduce_type=None):
        """Transform a series of samples into scattering domain.
        
        Arguments
        ---------
        samples: np.ndarray
            The input data to transform.
        
        reduce_type: str
            The reduction type (max, avg, med). 

        Returns
        -------
        features: list
            The features per layer of the scattering network.
        """
        # initialize empty lists for each layer
        features = [[] for _ in range(self.depth)]

        # transform each sample
        for sample in samples:
            scatterings = self.transform_sample(sample, reduce_type)
            for index, scattering in enumerate(scatterings):
                features[index].append(scattering)

        if cp.__name__ == "numpy": # if using numpy instead of numpy
            return [cp.array(feature) for feature in features]
        else:
            return [cp.array(feature).get() for feature in features]

    @property
    def depth(self):
        """Network depth."""
        return len(self.banks)



from setuptools import setup, find_packages

setup(
    name="scatseisnet",
    version="0.1.6",
    packages=find_packages(),
    python_requires=">=3.6, <4",
    install_requires=["astropy", "click", "fastcluster", "nmmn", "numpy", "obspy", "pandas", "parse", "seaborn", "sklearn", "tqdm"],
    entry_points={"console_scripts": ["scatseisnet = scatseisnet.cli.cli:main",],},
)



#!/usr/bin/env python
# coding=utf-8
"""Deep scattering transform clustering on segmented time series."""

__all__ = ["cluster", "linkage", "fcluster", "display"]

from . import cluster
from . import display
from .cluster import linkage, fcluster



# coding=utf-8
"""Deep scattering transform clustering on segmented time series."""

__all__ = [
    "show_inventory",
    "show_scatterings",
    "show_features",
    "dendrogram",
    "show_waveforms",
]

from .inventory import show_inventory
from .scatterings import show_scatterings
from .features import show_features
from .linkage import dendrogram
from .waveforms import show_waveforms



# coding: utf-8
"""Signal utilities.

Authors: Leonard Seydoux and Randall Balestriero
Email: leonard.seydoux@univ-grenoble-alpes.fr
Date: May, 2021
"""

import numpy as np


def segment(x, window_size, stride=None):
    """Segment a given array into (possibly overlapping) windows.

    Arguments
    ---------
    x: :class:`np.ndarray`
        The input array to segment.
    
    window_size: int
        The size of the sliding window.
    
    stride: int, optional
        The number of bins to slide the window with.
    
    Yields
    ------
    The segmented array with shape.
    """
    bins = x.shape[-1]
    index = 0
    stride = window_size if stride is None else stride
    while (index + window_size) <= bins:
        yield x[..., index : index + window_size]
        index += stride


def segmentize(x, window_size, stride=None):
    """Segment a given array into (possibly overlapping) windows.

    Arguments
    ---------
    x: :class:`np.ndarray`
        The input array to segment.
    
    window_size: int
        The size of the sliding window.
    
    stride: int, optional
        The number of bins to slide the window with.
    
    Returns
    -------
    The segmented array with shape (n_windows, n_channels, n_times).
    """
    return np.array([x for x in segment(x, window_size, stride)])


def pool(x, reduce_type="avg"):
    """Pooling operation performed on the last axis.

    Arguments
    ---------
    data: symjax.tensor or np.ndarray
        The input data to pool.

    Keyword arguments
    -----------------
    reduce_type: str
        The reducing operation (default: avg).

    Returns
    -------
    data_pooled: symjax.tensor
        The data pooled with same shape of input data minus last dimension.
    """
    if reduce_type == "avg":
        return x.mean(axis=-1)
    if reduce_type == "max":
        return x.max(axis=-1)
    if reduce_type is None:
        return x


def reshape_features(features, net):
    """Extract features from vector features in a single window."""
    # Extract dimensions
    n_filters_per_bank = [bank.shape[0] for bank in net.banks]
    n_features_per_layer = np.cumprod(n_filters_per_bank)
    n_features_per_channel = n_features_per_layer.sum()
    n_channels = features.shape[0] // n_features_per_channel

    # Loop over layers
    reshaped_features = list()
    start = 0
    for layer, n_features in enumerate(n_features_per_layer):
        end = start + n_channels * n_features
        feature = features[start:end]
        feature = feature.reshape(n_channels, *n_filters_per_bank[: layer + 1])
        reshaped_features.append(feature)
        start = end
    return reshaped_features


def normalize_features(features):
    """Normalized higher-order scattering coefficients."""
    n_layers = len(features)
    for layer in range(n_layers - 1):
        for channel in range(features[0].shape[0]):
            features[layer + 1][channel] /= (
                features[layer][channel][:, None] + 1e-5
            )
            # THIS IS FOR TESTING FULL NORM
            # for index, feature in enumerate(features[layer + 1][channel].T):
            #     features[layer + 1][channel][:, index] = (
            #         feature - feature.min()
            #     ) / (feature.max() - feature.min())

            # FULL NORM WITH SUM (PROBA)
            for index, _ in enumerate(features[layer + 1][channel].T):
                features[layer + 1][channel][:, index] /= np.sqrt(
                    features[layer + 1][channel][:, index].sum()
                )

    return features


def vectorize_features(features):
    """Extract features from vector features in a single window."""
    return np.hstack([f.reshape(-1) for f in features])



# coding: utf-8
"""Deep scattering transform on segmented time series.

Authors: Leonard Seydoux and Randall Balestriero
Email: leonard.seydoux@univ-grenoble-alpes.fr
Date: May, 2021
"""


import argparse
import click
import glob
import json
import logging
import numpy as np
import os

from pathlib import Path

FILE_ARGUMENTS = "arguments.json"


def stdout(message, value):
    """Format message with click echo and style.
    
    Arguments
    ---------
    message: str
        The message to use as a base string with {} variables.
    value: str
        The values to put in place of the {}.
    """
    # Happend curly brackets if not
    if "{" not in message:
        message = message + " {}"

    # Turn into tuple for allowing several values
    if not isinstance(value, tuple):
        values = (value,)
    else:
        values = value

    # Format
    values = [click.style(value, bold=True) for value in values]
    message = message.format(*values)

    # Echo
    click.echo(message)
    pass


def mkdir(dirpath):
    """Create directory.
    
    Arguments:
    ----------
    dirpath: str
        Path to the directory to create.
    """
    if not os.path.exists(dirpath):
        Path(dirpath).mkdir(parents=True, exist_ok=True)
        stdout("Created directory", dirpath)
    else:
        stdout("Using existing directory", dirpath)
    pass


def mkdirs(*args):
    """Create directories.
    
    Arguments:
    ----------
    dirpaths: tuple or list
        Paths to the directories to create.
    """
    for dirpath in args:
        mkdir(dirpath)


def parse(init=False):
    """Command-line argument parser.

    The usage and help for every command-line arguments for the main program is
    avaialbe from the main running script:

        python main.py -h

    assuming that the main.py script implements at least the following lines:

        import scat
        scat.parse()

    Keyword arguments
    -----------------
    readonly: bool
        A safe switch to use in order to prevent from erasing the data.

    Returns
    -------
    args: :class:`scat.argparse.Namespace`
        The parsed arguments.

    reader: module
        The imported data reading module.
    """

    # Instanciate parser with description and module's docstring.
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    # Logging level.
    parser.add_argument(
        "--log",
        default="INFO",
        metavar="LEVEL",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="""Logging level. The level name can either be NOTSET, DEBUG,
        INFO, WARNING (default), ERROR or CRITICAL. Please visit
        https://docs.python.org/3/howto/logging.html for further info.""",
    )

    # Output directory.
    parser.add_argument(
        "--outdir",
        metavar="PATH",
        type=str,
        default="out/case",
        help="""Specify output directory path. The directory should exist
        already.""",
    )

    # Window duration
    parser.add_argument(
        "--segment",
        metavar="DURATION",
        type=float,
        default=1800.0,
        help="""Segment duration in seconds. A default window duration of 20
        seconds is set, but this argument should be considered as a required
        argument. This segment also defines the pooling size, since the pooling
        is performed over the full segment duration.""",
    )

    # Window step
    parser.add_argument(
        "--step",
        metavar="DURATION",
        type=float,
        default=900.0,
        help="""Segment step in seconds. This defines the time interval between
        the starting time of two consecutive segments. The clustering
        performances with respect to the step is still under debate. for
        cleaner results, the step should be set with the segment duration.""",
    )

    # Number of octaves at each layer
    parser.add_argument(
        "--octaves",
        type=int,
        default=[7, 12],
        nargs="+",
        metavar="J",
        help="""Number of octaves for each scattering layer. The number of
        octaves define the freequency extent of each layer. The number of
        octaves is defined from the Nyquist frequency. The scattering network
        depth depends on the length of this list.""",
    )

    # Number of filter per octaves at each layer
    parser.add_argument(
        "--resolution",
        type=int,
        default=[6, 1],
        nargs="+",
        metavar="Q",
        help="""Number of wavelets per octaves for each scattering layer. This
        define the frequency resolution of each scattering layer, and
        consequently, the representation density. Note that the length of this
        argument must be the same that the number of octaves.""",
    )

    # Wavelet banks quality factors
    parser.add_argument(
        "--quality",
        type=int,
        default=[1, 1],
        nargs="+",
        metavar="Qc",
        help="""Wavelet bank qulity factor. This defines the ratio bewteen
        the center frequency qnd the frequency bandwitdh, and therefore
        represents the selectivity of the filter, or quality. It allows to
        choose the density level of each representation.""",
    )

    # Pooling type
    parser.add_argument(
        "--pooling",
        metavar="TYPE",
        type=str,
        default="max",
        choices=["max", "avg"],
        help="""Pooling reduction operation. The pooling is performed on a
        the duration of a full segment. By default, the maximum pooling
        is performed (choices: max or avg).""",
    )

    # Waveform inventory
    parser.add_argument(
        "--inventory",
        metavar="PATH",
        type=str,
        default="INVENTORY",
        help="""Waveform inventory to read with obspy.""",
    )

    # Features files
    parser.add_argument(
        "--file_features",
        metavar="PATH",
        type=str,
        default="out/case/features/*.npz",
        help="""Path to features, where the wildcard is replaced by the
        tag.""",
    )

    # Network file
    parser.add_argument(
        "--file_network",
        metavar="PATH",
        type=str,
        default="out/case/models/network.pickle",
        help="""Path to network pickle file for saving.""",
    )

    if parser.prog == "make_latent.py":

        parser.add_argument(
            "dim",
            metavar="DIM",
            type=int,
            default=10,
            help="""Latent space dimensions.""",
        )

        parser.add_argument(
            "--file_latent",
            metavar="PATH",
            type=str,
            default="out/case/latent/latent.npz",
            help="""Path to latent file for saving.""",
        )

        parser.add_argument(
            "--file_reduction",
            metavar="PATH",
            type=str,
            default="out/case/model/reduce.pickle",
            help="""Path to reduction model pickle file for saving.""",
        )

        parser.add_argument(
            "--normalize",
            metavar="NORM",
            type=int,
            default=0,
            help="""If 1, the higher-order scatterings are normalized.""",
        )

    if parser.prog == "make_cluster.py":

        parser.add_argument(
            "--depth",
            metavar="depth",
            type=int,
            default=40,
            help="""Number of dendrogram splits.""",
        )

        parser.add_argument(
            "--threshold",
            metavar="THRESHOLD",
            type=float,
            default=0.5,
            help="""Threshold for cluster definition.""",
        )

        parser.add_argument(
            "--distance",
            metavar="distance",
            type=str,
            default="ward",
            help="""Method to calculate clusters.""",
        )

        parser.add_argument(
            "--file_leaves",
            metavar="FILE_LEAVES",
            type=str,
            default="out/case/model/leaves.pickle",
            help="""Save file for leaves.""",
        )

        parser.add_argument(
            "--file_clustering",
            metavar="FILE_CLUSTERING",
            type=str,
            default="out/case/model/clusters.pickle",
            help="""Save file for clusters.""",
        )

        parser.add_argument(
            "--file_latent",
            metavar="PATH",
            type=str,
            default="out/case/latent/latent.npz",
            help="""Path to latent file for saving.""",
        )

        parser.add_argument(
            "--file_linkage",
            metavar="PATH",
            type=str,
            default="out/case/cluster/linkage.pickle",
            help="""Path to linkage file for saving.""",
        )

        parser.add_argument(
            "--linkage",
            metavar="linkage",
            type=int,
            default=1,
            help="""Recalculate linkage.""",
        )
    # Parse command-line arguments
    args = parser.parse_args()

    # Set logging
    logging.basicConfig(format="[%(levelname)s] %(message)s", level=args.log)
    logging.info(f"Logging level set to {args.log}")

    if init == True:
        args = parser.parse_args()
        save_arguments(args)
        return args

    else:
        return load_arguments(args)


def save_arguments(args, skip=["log"]):
    """Save arguments.

    The input arguments out of the parser function are saved into a JSON file
    format in order to ensure readability by the user.

    Arguments
    ---------
    args: :class:`scat.argparse.Namespace`
        The argument to save.

    skip: list
        A list of arguments not to save. Useful for avoiding saved arguments
        to overwrite new arguments in later runs like log or mode.
    """
    # Filename
    json_filename = os.path.join(args.outdir, FILE_ARGUMENTS)

    # Skip arguments from the list that are not to be saved
    args_dict = args.__dict__.copy()
    for arg_skip in skip:
        args_dict.pop(arg_skip)

    # Save
    with open(json_filename, "w") as json_file:
        json.dump(args_dict, json_file, indent=4)

    # Logging checkpoint
    logging.info("Saved arguments at {}".format(json_filename))
    pass


def load_arguments(args=argparse.Namespace(), filename=None):
    """Load arguments.

    This function allows to update the arguments pre-loaded by the parser from
    a JSON file previously saved with the save_arguments function. It allows to
    recover a set of arguments of a previous run.

    Arguments
    ---------
    args: :class:`argparse.Namespace`
        The default arguments to overwrite.

    Returns
    -------
    args: :class:`argparse.Namespace`
        The updated arguments.
    """
    # Filename
    if filename is not None:
        json_filename = filename
    else:
        json_filename = os.path.join(args.outdir, FILE_ARGUMENTS)

    # Read and update arguments
    with open(json_filename, "r") as json_file:
        args.__dict__.update(json.load(json_file))

    # Logging checkpoint
    logging.info("Loaded arguments from {}".format(json_filename))
    logging.debug("Arguments list {}".format(args))
    return args


def load_features(path):
    """Read scattering coefficients and return design matrix."""
    # init
    x = list()
    t = list()

    # loop over available npz files
    for datafile in sorted(glob.glob(path)):
        data = np.load(datafile)
        features = [data[key] for key in data if "features" in key]

        # reshape and stack
        if features[0].shape[0]:
            features = [feat.reshape(feat.shape[0], -1) for feat in features]
            x.append(np.hstack(features))
            t.append(data["times"])

    return np.vstack(x), np.hstack(t)


def load_feature_file(path):
    """Read scattering coefficients and return design matrix."""
    data = np.load(path)
    features = [data[key] for key in data if "features" in key]
    if features[0].shape[0]:
        features = [feat.reshape(feat.shape[0], -1) for feat in features]
        x = np.hstack(features)
        t = data["times"]
    return x, t
